//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/DistributionItems.cpp
//! @brief     Implements class DistributionItem and several subclasses
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Descriptor/DistributionItems.h"
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Support/XML/Serialize.h"
#include "Param/Distrib/Distributions.h"
#include "Param/Distrib/RangedDistributions.h"
#include <cmath>

namespace {
namespace Tag {

const QString NumberOfSamples("NumberOfSamples");
const QString SigmaFactor("SigmaFactor");
const QString Mean("Mean");
const QString Minimum("Minimum");
const QString Maximum("Maximum");
const QString HWHM("HWHM");
const QString StandardDeviation("StandardDeviation");
const QString Median("Median");
const QString ScaleParameter("ScaleParameter");
const QString Sigma("Sigma");
const QString Center("Center");
const QString LeftWidth("LeftWidth");
const QString MiddleWidth("MiddleWidth");
const QString RightWidth("RightWidth");

} // namespace Tag
} // namespace

using std::variant;

DistributionItem::DistributionItem() {}

void DistributionItem::setLimits(const RealLimits& limits)
{
    m_limits = limits;
}

void DistributionItem::initSigmaFactor()
{
    m_sigmaFactor.init("Sigma factor", "", 2.0, Unit::unitless, "sigmaFactor");
}

bool DistributionItem::hasSigmaFactor() const
{
    return m_sigmaFactor.isInitialized();
}

void DistributionItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwValue(s, Tag::NumberOfSamples, m_numberOfSamples);

    if (hasSigmaFactor())
        Serialize::rwProperty(s, Tag::SigmaFactor, m_sigmaFactor);

    // m_limits and a changed unit (with setUnit) will not be serialized here. They have to be set
    // again by the owner of DistributionItem after reading it
}

// --------------------------------------------------------------------------------------------- //

SymmetricResolutionItem::SymmetricResolutionItem(double mean, int decimals,
                                                 const QString& meanLabel)
{
    m_mean.init(meanLabel, "", mean, Unit::unitless, decimals, RealLimits::limitless(), "mean");
}

void SymmetricResolutionItem::setUnit(const variant<QString, Unit>& unit)
{
    m_mean.setUnit(unit);
}

void SymmetricResolutionItem::setLimits(const RealLimits& limits)
{
    DistributionItem::setLimits(limits);
    m_mean.setLimits(limits);
}

void SymmetricResolutionItem::setMeanDecimals(uint d)
{
    m_mean.setDecimals(d);
}

void SymmetricResolutionItem::serialize(Streamer& s)
{
    DistributionItem::serialize(s);
    Serialize::rwProperty(s, Tag::Mean, m_mean);
}

// --------------------------------------------------------------------------------------------- //

DistributionNoneItem::DistributionNoneItem()
    : SymmetricResolutionItem(0.1, 3, "Value")
{
}

std::unique_ptr<IDistribution1D> DistributionNoneItem::createDistribution(double) const
{
    return nullptr;
}

std::unique_ptr<IRangedDistribution> DistributionNoneItem::createIRangedDistribution(double) const
{
    return nullptr;
}

double DistributionNoneItem::deviation(double) const
{
    return 0.0;
}

void DistributionNoneItem::initDistribution(double value)
{
    setMean(value);
}

DoublePropertyRefs DistributionNoneItem::distributionValues(bool withMean)
{
    return withMean ? DoublePropertyRefs{m_mean} : DoublePropertyRefs{};
}

// --------------------------------------------------------------------------------------------- //

DistributionGateItem::DistributionGateItem()
{
    m_minimum.init("Min", "", 0.0, Unit::unitless, 3 /* decimals */, RealLimits::limitless(),
                   "min");
    m_maximum.init("Max", "", 1.0, Unit::unitless, 3 /* decimals */, RealLimits::limitless(),
                   "max");
}

void DistributionGateItem::setUnit(const variant<QString, Unit>& unit)
{
    m_minimum.setUnit(unit);
    m_maximum.setUnit(unit);
}

std::unique_ptr<IDistribution1D> DistributionGateItem::createDistribution(double scale) const
{
    return std::make_unique<DistributionGate>(scale * m_minimum.get(), scale * m_maximum.get());
}

void DistributionGateItem::initDistribution(double value)
{
    double sigma(0.1 * std::abs(value));
    if (sigma == 0.0)
        sigma = 0.1;
    m_minimum.set(value - sigma);
    m_maximum.set(value + sigma);
}

void DistributionGateItem::setRange(double min, double max)
{
    m_minimum.set(min);
    m_maximum.set(max);
}

void DistributionGateItem::serialize(Streamer& s)
{
    DistributionItem::serialize(s);
    Serialize::rwProperty(s, Tag::Minimum, m_minimum);
    Serialize::rwProperty(s, Tag::Maximum, m_maximum);
}

DoublePropertyRefs DistributionGateItem::distributionValues(bool /*withMean*/)
{
    return {m_minimum, m_maximum};
}

// --------------------------------------------------------------------------------------------- //

DistributionLorentzItem::DistributionLorentzItem()
    : SymmetricResolutionItem(1.0)
{
    initSigmaFactor();

    m_hwhm.init("HWHM", "", 1.0, Unit::unitless, "hwhm");
}

void DistributionLorentzItem::setUnit(const variant<QString, Unit>& unit)
{
    SymmetricResolutionItem::setUnit(unit);
    m_hwhm.setUnit(unit);
}

std::unique_ptr<IDistribution1D> DistributionLorentzItem::createDistribution(double scale) const
{
    return std::make_unique<DistributionLorentz>(scale * m_mean.get(), scale * m_hwhm.get());
}

std::unique_ptr<IRangedDistribution>
DistributionLorentzItem::createIRangedDistribution(double scale) const
{
    return std::make_unique<RangedDistributionLorentz>(numberOfSamples(), m_sigmaFactor.get(),
                                                       m_limits.scaledLimits(scale));
}

double DistributionLorentzItem::deviation(double scale) const
{
    return m_hwhm * scale;
}

void DistributionLorentzItem::initDistribution(double value)
{
    double hw(0.1 * std::abs(value));
    if (hw == 0.0)
        hw = 0.1;

    setMean(value);
    setHwhm(hw);
    m_hwhm.setLimits(RealLimits::lowerLimited(0.0));
}

void DistributionLorentzItem::serialize(Streamer& s)
{
    SymmetricResolutionItem::serialize(s);
    Serialize::rwProperty(s, Tag::HWHM, m_hwhm);
}

DoublePropertyRefs DistributionLorentzItem::distributionValues(bool withMean)
{
    return withMean ? DoublePropertyRefs{m_mean, m_hwhm, m_sigmaFactor}
                    : DoublePropertyRefs{m_hwhm, m_sigmaFactor};
}

// --------------------------------------------------------------------------------------------- //

DistributionGaussianItem::DistributionGaussianItem()
    : SymmetricResolutionItem(1.0)
{
    initSigmaFactor();

    m_standardDeviation.init("StdDev", "", 1.0, Unit::unitless, 3 /* decimals */,
                             RealLimits::lowerLimited(0.0), "stdDev");
}

void DistributionGaussianItem::setUnit(const variant<QString, Unit>& unit)
{
    SymmetricResolutionItem::setUnit(unit);
    m_standardDeviation.setUnit(unit);
}

std::unique_ptr<IDistribution1D> DistributionGaussianItem::createDistribution(double scale) const
{
    return std::make_unique<DistributionGaussian>(scale * m_mean.get(),
                                                  scale * m_standardDeviation.get());
}

std::unique_ptr<IRangedDistribution>
DistributionGaussianItem::createIRangedDistribution(double scale) const
{
    return std::make_unique<RangedDistributionGaussian>(m_numberOfSamples, m_sigmaFactor.get(),
                                                        m_limits.scaledLimits(scale));
}

double DistributionGaussianItem::deviation(double scale) const
{
    return m_standardDeviation.get() * scale;
}

void DistributionGaussianItem::initDistribution(double value)
{
    double stddev(0.1 * std::abs(value));
    if (stddev == 0.0)
        stddev = 0.1;

    setMean(value);
    setStandardDeviation(stddev);
}

void DistributionGaussianItem::serialize(Streamer& s)
{
    SymmetricResolutionItem::serialize(s);
    Serialize::rwProperty(s, Tag::StandardDeviation, m_standardDeviation);
}

DoublePropertyRefs DistributionGaussianItem::distributionValues(bool withMean)
{
    return withMean ? DoublePropertyRefs{m_mean, m_standardDeviation, m_sigmaFactor}
                    : DoublePropertyRefs{m_standardDeviation, m_sigmaFactor};
}

// --------------------------------------------------------------------------------------------- //

DistributionLogNormalItem::DistributionLogNormalItem()
{
    initSigmaFactor();

    m_median.init("Median", "", 1.0, Unit::unitless, "median");
    m_scaleParameter.init("ScaleParameter", "", 1.0, Unit::unitless, 3 /* decimals */,
                          RealLimits::lowerLimited(0.0), "scalePar");
}

void DistributionLogNormalItem::setUnit(const variant<QString, Unit>& unit)
{
    m_median.setUnit(unit);
}

std::unique_ptr<IDistribution1D> DistributionLogNormalItem::createDistribution(double scale) const
{
    return std::make_unique<DistributionLogNormal>(scale * m_median.get(), m_scaleParameter.get());
}

void DistributionLogNormalItem::initDistribution(double value)
{
    double scale(0.1 * std::abs(value));
    if (scale == 0.0)
        scale = 0.1;

    setMedian(value);
    setScaleParameter(scale);
}

void DistributionLogNormalItem::serialize(Streamer& s)
{
    DistributionItem::serialize(s);
    Serialize::rwProperty(s, Tag::Median, m_median);
    Serialize::rwProperty(s, Tag::ScaleParameter, m_scaleParameter);
}

DoublePropertyRefs DistributionLogNormalItem::distributionValues(bool /*withMean*/)
{
    return {m_median, m_scaleParameter, m_sigmaFactor};
}

// --------------------------------------------------------------------------------------------- //

DistributionCosineItem::DistributionCosineItem()
    : SymmetricResolutionItem(1.0)
{
    initSigmaFactor();
    m_sigma.init("Sigma", "", 1.0, Unit::unitless, 3 /* decimals */, RealLimits::lowerLimited(0.0),
                 "sigma");
}

void DistributionCosineItem::setUnit(const variant<QString, Unit>& unit)
{
    SymmetricResolutionItem::setUnit(unit);
    m_sigma.setUnit(unit);
}

std::unique_ptr<IDistribution1D> DistributionCosineItem::createDistribution(double scale) const
{
    return std::make_unique<DistributionCosine>(scale * m_mean.get(), scale * m_sigma.get());
}

std::unique_ptr<IRangedDistribution>
DistributionCosineItem::createIRangedDistribution(double scale) const
{
    return std::make_unique<RangedDistributionCosine>(m_numberOfSamples, m_sigmaFactor.get(),
                                                      m_limits.scaledLimits(scale));
}

double DistributionCosineItem::deviation(double scale) const
{
    return m_sigma.get() * scale;
}

void DistributionCosineItem::initDistribution(double value)
{
    double sigma(0.1 * std::abs(value));
    if (sigma == 0.0)
        sigma = 0.1;

    setMean(value);
    setSigma(sigma);
}

void DistributionCosineItem::serialize(Streamer& s)
{
    SymmetricResolutionItem::serialize(s);
    Serialize::rwProperty(s, Tag::Sigma, m_sigma);
}

DoublePropertyRefs DistributionCosineItem::distributionValues(bool withMean)
{
    return withMean ? DoublePropertyRefs{m_mean, m_sigma, m_sigmaFactor}
                    : DoublePropertyRefs{m_sigma, m_sigmaFactor};
}

// --------------------------------------------------------------------------------------------- //

DistributionTrapezoidItem::DistributionTrapezoidItem()
{
    m_center.init("Center", "", 1.0, Unit::unitless, 3 /* decimals */, RealLimits::limitless(),
                  "center");
    m_leftWidth.init("LeftWidth", "", 1.0, Unit::unitless, "left");
    m_middleWidth.init("MiddleWidth", "", 1.0, Unit::unitless, "middle");
    m_rightWidth.init("RightWidth", "", 1.0, Unit::unitless, "right");
}

void DistributionTrapezoidItem::setUnit(const variant<QString, Unit>& unit)
{
    m_center.setUnit(unit);
    m_leftWidth.setUnit(unit);
    m_middleWidth.setUnit(unit);
    m_rightWidth.setUnit(unit);
}

std::unique_ptr<IDistribution1D> DistributionTrapezoidItem::createDistribution(double scale) const
{
    return std::make_unique<DistributionTrapezoid>(
        scale * m_center.get(), scale * m_leftWidth.get(), scale * m_middleWidth.get(),
        scale * m_rightWidth.get());
}

void DistributionTrapezoidItem::initDistribution(double value)
{
    double width(0.1 * std::abs(value));
    if (width == 0.0)
        width = 0.1;
    setCenter(value);
    setLeftWidth(width);
    setMiddleWidth(width);
    setRightWidth(width);
}

void DistributionTrapezoidItem::serialize(Streamer& s)
{
    DistributionItem::serialize(s);
    Serialize::rwProperty(s, Tag::Center, m_center);
    Serialize::rwProperty(s, Tag::LeftWidth, m_leftWidth);
    Serialize::rwProperty(s, Tag::MiddleWidth, m_middleWidth);
    Serialize::rwProperty(s, Tag::RightWidth, m_rightWidth);
}

DoublePropertyRefs DistributionTrapezoidItem::distributionValues(bool withMean)
{
    return withMean ? DoublePropertyRefs{m_center, m_leftWidth, m_middleWidth, m_rightWidth}
                    : DoublePropertyRefs{m_leftWidth, m_middleWidth, m_rightWidth};
}
