//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/DoubleDescriptor.h
//! @brief     Defines class DoubleDescriptor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_DOUBLEDESCRIPTOR_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_DOUBLEDESCRIPTOR_H

#include "Fit/Param/RealLimits.h"
#include "GUI/Support/Type/Unit.h"
#include <QList>
#include <QString>
#include <functional>
#include <variant>

//! Describes properties of a double value which are necessary to allow GUI representation, editing
//! the value, undo/redo, unit conversion.
//!
//! By using this class, the underlying data scheme is hidden from the user of the data. The
//! underlying implementation can be a simple double member, or any other construction to hold a
//! double value. It could also describe a value which is not even represented by an existing
//! double, but is only a calculated value. In this case it might apply that there is no setter
//! possible, which is still allowed as long as care is taken not trying to call the setter.
//!
//! Note that this class does not provide (*implement*) a double value, but *provide information*
//! about a double value. For implementing a double value, please see DoubleProperty.
//!
//! For easy UI creation, there are functions like GUI:Util::createSpinBox() which take a descriptor
//! and fully initialize the created spin box.
//!
//! \sa DoubleProperty
//! \sa GUI:Util::createSpinBox()
//!
class DoubleDescriptor {
private: // private as long as path initialization is not included in params (to be done after
         // SessionItem migration)
    //! Operates on a double value (e.g a member variable).
    DoubleDescriptor(const QString& label, const QString& tooltip, int decimals,
                     const RealLimits& limits, double* var,
                     const std::variant<QString, Unit>& unit);

    //! Operates on a double value (e.g a member variable).
    //! Decimals is set to 3, limits is set to nonnegative
    DoubleDescriptor(const QString& label, const QString& tooltip, double* var,
                     const std::variant<QString, Unit>& unit);

public:
    //! Operates on any kind of storage (e.g. session items), by using setter/getter methods
    //! decimals is set to 3, limits is set to nonnegative
    DoubleDescriptor(const QString& label, const QString& tooltip,
                     std::function<void(double)> setter, std::function<double()> getter,
                     const std::variant<QString, Unit>& unit);

    //! Operates on any kind of storage (e.g. session items), by using setter/getter methods
    DoubleDescriptor(QString label, QString tooltip, int decimals, const RealLimits& limits,
                     std::function<void(double)> setter, std::function<double()> getter,
                     const std::variant<QString, Unit>& unit);

    DoubleDescriptor(const QString& label, const double* var,
                     const std::variant<QString, Unit>& unit);

    DoubleDescriptor() = default;

public:
    //! Return the current value of the handled parameter.
    operator double() const;

    QString label;     //!< A label text (short, no trailing colon)
    QString tooltip;   //!< Tooltip text
    int decimals = 0;  //!< numbers of decimals to be shown in an edit control
    RealLimits limits; //!< Limits of the value.
    std::function<void(double)> set = nullptr;         //!< function to set the value
    std::function<double()> get = nullptr;             //!< function to get the current value
    std::variant<QString, Unit> unit = Unit::unitless; //!< Unit of the value (internal unit only!)
    std::function<QString()> path =
        nullptr; //!< Path describing this value. Used e.g. for undo/redo
};

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_DOUBLEDESCRIPTOR_H
